#include "PHASIC++/Process/Virtual_ME2_Base.H"
#include "AddOns/MCFM/MCFM_Wrapper.H"
 
namespace MCFM {
  class MCFM_qqb_WW: public PHASIC::Virtual_ME2_Base {
  private:
    int     m_pID;
    double  m_aqed,m_cplcorr, m_normcorr;
    bool    m_change_order;
    bool    m_four_lepton;
    double *p_p, *p_msqv;
   
    double CallMCFM(const int & i,const int & j);
  public:
    MCFM_qqb_WW(const int & pID,const PHASIC::Process_Info& pi,
		const ATOOLS::Flavour_Vector& flavs, bool change_order,
		bool four_lepton);
    ~MCFM_qqb_WW();
    void Calc(const ATOOLS::Vec4D_Vector& momenta);
    double Eps_Scheme_Factor(const ATOOLS::Vec4D_Vector& mom);
  };

}// end of namespace MCFM

extern "C" { 
  void qqb_ww_v_(double *p,double *msqv); 
}

#include "MODEL/Main/Model_Base.H"
#include "ATOOLS/Org/Run_Parameter.H"

using namespace MCFM;
using namespace PHASIC;
using namespace ATOOLS;

MCFM_qqb_WW::MCFM_qqb_WW(const int & pID,const PHASIC::Process_Info& pi,
			 const Flavour_Vector& flavs, bool change_order,
			 bool four_lepton):
  Virtual_ME2_Base(pi,flavs), m_pID(pID),
  m_aqed(MODEL::s_model->ScalarFunction(std::string("alpha_QED"))),
  m_cplcorr(ATOOLS::sqr(4.*M_PI*m_aqed/ewcouple_.esq)),
  m_normcorr(4.*9./qcdcouple_.ason2pi),
  m_change_order(change_order),
  m_four_lepton(four_lepton)
{
  rpa->gen.AddCitation
    (1,"The NLO matrix elements have been taken from MCFM \\cite{Campbell:1999ah}.");
  p_p = new double[4*MCFM_NMX];
  p_msqv = new double[sqr(2*MCFM_NF+1)];
  m_drmode=m_mode=1;
}

MCFM_qqb_WW::~MCFM_qqb_WW()
{
  delete [] p_p;
  delete [] p_msqv;
}

double MCFM_qqb_WW::CallMCFM(const int & i,const int & j) {
  switch (m_pID) {
  case 61: qqb_ww_v_(p_p,p_msqv); break;
  }
  return p_msqv[mr(i,j)];
}

void MCFM_qqb_WW::Calc(const Vec4D_Vector &p)
{
  double corrfactor(m_cplcorr*m_normcorr);

  for (int n(0);n<2;++n) GetMom(p_p,n,-p[n]);
  if (m_change_order){
    GetMom(p_p,2,p[4]);
    GetMom(p_p,3,p[5]);
    GetMom(p_p,4,p[2]);
    GetMom(p_p,5,p[3]);
  }
  else if (m_four_lepton)  {
    GetMom(p_p,2,p[3]);
    GetMom(p_p,3,p[4]);
    GetMom(p_p,4,p[2]);
    GetMom(p_p,5,p[5]); 
  }

  else
    for (size_t n(2);n<p.size();++n) GetMom(p_p,n,p[n]);
  
  long int i(m_flavs[0]), j(m_flavs[1]);
  if (i==21) { i=0; }
  if (j==21) { j=0; }
  scale_.musq=m_mur2;
  scale_.scale=sqrt(scale_.musq);

  epinv_.epinv=epinv2_.epinv2=0.0;
  double res(CallMCFM(i,j)  * corrfactor);
  epinv_.epinv=1.0;
  double res1(CallMCFM(i,j) * corrfactor);
  epinv2_.epinv2=1.0;
  double res2(CallMCFM(i,j) * corrfactor);

  m_res.Finite() = res;
  m_res.IR()     = (res1-res);
  m_res.IR2()    = (res2-res1);
}

double MCFM_qqb_WW::Eps_Scheme_Factor(const ATOOLS::Vec4D_Vector& mom)
{
  return 4.*M_PI;
}

extern "C" { void chooser_(); }

DECLARE_VIRTUALME2_GETTER(MCFM_qqb_WW,"MCFM_qqb_WW")
Virtual_ME2_Base *ATOOLS::Getter
<Virtual_ME2_Base,Process_Info,MCFM_qqb_WW>::
operator()(const Process_Info &pi) const
{
  DEBUG_FUNC("");
  if (MODEL::s_model->Name()!=std::string("SM")
      && MODEL::s_model->Name()!=std::string("SM+AGC")) return NULL;
  if (pi.m_loopgenerator!="MCFM")                       return NULL;
  if (pi.m_fi.m_nloewtype!=nlo_type::lo)                return NULL;
  if (!pi.m_fi.m_nloqcdtype&nlo_type::loop)             return NULL;
  Flavour_Vector fl(pi.ExtractFlavours());
  // check for right flavour structure:
  // - two incoming quarks + 4 particle FS
  // - assume two propagators - check for alternatives (i.e. 4 lepton FS) later
  // - have W/Z
  if (!(fl[0].IsQuark() && fl[1].IsQuark()))            return NULL;
  if (fl.size()!=6)                                     return NULL;
  if (!(fl[2].IsLepton() && fl[3].IsLepton() &&
	fl[4].IsLepton() && fl[5].IsLepton())) {
    msg_Error()<<"Error in "<<METHOD<<":\n"
	       <<"   WW production for lepton/neutrino final states only.\n";
    THROW(fatal_error,"Not yet working."); 
  }
  int pID(0);
  bool change_order(false);
  bool four_lepton(false);
  if (pi.m_fi.m_ps.size()==2 &&
      pi.m_fi.m_ps[0].m_fl[0].Kfcode()==24 && pi.m_fi.m_ps[0].m_fl[1].Kfcode()==24) {
    if (Flavour(kf_b).Yuk()>0.) {
      msg_Error()<<"Error in "<<METHOD<<":\n"
		 <<"   Must switch off Yukawa couplings of b quarks.\n";
      THROW(fatal_error,"Wrong model assupmtions."); 
    }
    if (Flavour(kf_t).IsOn() || 
	Flavour(kf_b).IsOn() && !(Flavour(kf_b).Mass())) {
      msg_Error()<<"Error in "<<METHOD<<":\n"
		 <<"   Must switch off the top in the t-channel.\n";
      THROW(fatal_error,"Wrong model assupmtions."); 
    }
    pID = 61;
  }
  else if (fl[2].IsDowntype() && fl[3].IsUptype()) {
    if (MODEL::s_model->Name()==std::string("SM+AGC")) {
      msg_Error()<<"Error in "<<METHOD<<":\n"
		 <<"   Must have explicit boson decays for anomalous couplings.\n";
      THROW(fatal_error,"Not working."); 
    }
    if (Flavour(kf_b).Yuk()>0.) {
      msg_Error()<<"Error in "<<METHOD<<":\n"
		 <<"   Must switch off Yukawa couplings of b quarks.\n";
      THROW(fatal_error,"Wrong model assupmtions."); 
    }
    if (Flavour(kf_t).IsOn() || 
	Flavour(kf_b).IsOn() && !(Flavour(kf_b).Mass())) {
      msg_Error()<<"Error in "<<METHOD<<":\n"
		 <<"   Must switch off the top in the t-channel.\n";
      THROW(fatal_error,"Wrong model assupmtions."); 
    }
    pID=61;
    four_lepton=true;
  }
  if (fl[2].IsDowntype() && !(four_lepton)) change_order=true;
  if (pID!=0) {
    if (nproc_.nproc>=0) {
      if (nproc_.nproc!=pID)
	THROW(not_implemented,
	      "Only one process class allowed when using MCFM");
    }
    nproc_.nproc=pID;
    if (pID=61) {
      if (four_lepton) zerowidth_.zerowidth=false;
      else             zerowidth_.zerowidth=true;
    }
    chooser_();
    msg_Info()<<"Initialise MCFM with nproc = "<<nproc_.nproc<<"\n";
    return new MCFM_qqb_WW(pID,pi,fl,change_order,four_lepton);
  }
  return NULL;
}

